#include <iostream>
#include "kd_tree.h"
#include "utilities.h"

#include <mpi.h>

using std::vector ;

using std::cerr ;
using std::cout ;
using std::endl ;

void parallel_random_points(vector<vector3d> &pnts, vector<int> &ids, int N,
                            MPI_Comm comm) {
  int i = 0 ;
  MPI_Comm_rank(comm,&i) ;
  int p = 0 ;
  MPI_Comm_size(comm,&p) ;

  int local_size = N/p + (i<(N%p)?1:0) ;
  vector<vector3d> vtmp(local_size) ;
  pnts.swap(vtmp) ;
  vector<int> itmp(local_size) ;
  ids.swap(itmp) ;

  if(i==0) {
    for(int j=0;j<local_size;++j) {
      double x = drand48() ;
      double y = drand48() ;
      double z = drand48() ;
      pnts[j] = vector3d(x,y,z) ;
      ids[j] = j ;
    }
    // Now that we assigned our processor values, compute remaining processors
    // values and send to them.
    int off = local_size ;
    for(int j=1;j<p;++j) {
      int psize = N/p + (i<(N%p)?1:0) ;
      vector<vector3d> data(psize) ;
      for(int k=0;k<psize;++k) {
        double x = drand48() ;
        double y = drand48() ;
        double z = drand48() ;
        data[k] = vector3d(x,y,z) ;
      }
      MPI_Send(&off,1,MPI_INT,j,0,comm) ;
      MPI_Send(&data[0],psize*3,MPI_DOUBLE,j,0,comm) ;
      off += psize ;
    }
  } else {
    int off = 0 ;
    MPI_Status stat ;
    MPI_Recv(&off,1,MPI_INT,0,0,comm,&stat) ;
    MPI_Recv(&pnts[0],local_size*3,MPI_DOUBLE,0,0,comm,&stat) ;
    for(int j=0;j<local_size;++j)
      ids[j] = off+j ;
  }
}
  

// Verify result by collecting the results and performing a serial match
void Verify(vector<vector3d> &target_pnts, vector<int> &target_ids,
            vector<vector3d> &search_pnts, vector<int> &close_pt) {

  // Verify by collecting all of the target points on all of the processors
  // and then using the naive algorithm.
  
  int numprocs = 0 ;
  int myid = 0 ;
  /* Get the number of processors and my processor identification */
  MPI_Comm_size(MPI_COMM_WORLD,&numprocs) ;
  MPI_Comm_rank(MPI_COMM_WORLD,&myid) ;


  int sz = target_pnts.size() ;
  vector<vector3d> collect_pnts(sz*numprocs) ;
  vector<int> collect_ids(sz*numprocs) ;
  MPI_Allgather(&target_pnts[0],sz*3,MPI_DOUBLE,
               &collect_pnts[0],sz*3,MPI_DOUBLE,
               MPI_COMM_WORLD) ;
  MPI_Allgather(&target_ids[0],sz,MPI_INT,
               &collect_ids[0],sz,MPI_INT,
               MPI_COMM_WORLD) ;

  int num_fail = 0 ;
  for(int i=0;i<search_pnts.size();++i) {
    double dist = 1e65 ;
    int id = -1 ;
    for(int j=0;j<collect_pnts.size();++j) {
      double dtmp = dist2(search_pnts[i],collect_pnts[j]) ;
      if(dist > dtmp) {
        dist = dtmp ;
        id = collect_ids[j] ;
      }
    }
    if(id != close_pt[i]) {
      num_fail++ ;
    }
  }

  int num_fail_tot = num_fail ;
  
  MPI_Reduce(&num_fail,&num_fail_tot,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD) ;
  if(0 == myid) {
    if(num_fail_tot != 0) {
      cout << "Verification failed for " << num_fail_tot
           << " cases out of " << search_pnts.size()*numprocs
           << "!!!!!!!!" << endl ;
    } else {
      cout << "Passed Verification" << endl ;
    }
  }
}



int main(int argc, char *argv[]) {

  chopsigs_() ;
  /* Initialize MPI */
  MPI_Init(&argc,&argv) ;

  bool verify = true ;
  
  int numprocs = 0 ;
  int myid = 0 ;
  /* Get the number of processors and my processor identification */
  MPI_Comm_size(MPI_COMM_WORLD,&numprocs) ;
  MPI_Comm_rank(MPI_COMM_WORLD,&myid) ;

  int N = 128*numprocs ;
  int M = 128*numprocs ;

  if(argc > 1)
    verify = false ;
  
  if(argc == 2) {
    N = atoi(argv[1]) ;
    M = N ;
  }
  if(argc == 3) 
    M = atoi(argv[2]) ;

  if(0 == myid)
    cout << "N = " << N << ", M = " << M << endl;

  // target_pnts is the given set of points
  // search_pnts is the query points
  vector<vector3d> target_pnts, search_pnts ;
  // target_id is an integer number identifier for the given set points
  // search_id is an integer number identifier for the search points
  vector<int> target_ids, search_ids ;

  // For testing, initialize target and search points to random positions
  // within a unit cube
  parallel_random_points(target_pnts,target_ids,N,MPI_COMM_WORLD) ;
  parallel_random_points(search_pnts,search_ids,M,MPI_COMM_WORLD) ;


  int lN = target_pnts.size() ;
  int lM = search_pnts.size() ;

  // close_pt is an array of target ids for each point in search_pnts array.
  // The target id will be set to the closest point in the target_pnts array.

  // For example the naive pseudocode for this would be {
  //  for(i in search points) {
  //    min_dist = inf ;
  //    min_id = -1 ;
  //    for(j in target_pnts) {
  //      if(min_dist > distance(search_pnts[i],target_pnts[j])) {
  //        min_dist = distance(search_pnts[i],target_pnts[j]) ;
  //        min_id = j ;
  //      }
  //    close_pt[i] = min_id ;
  //  }
  

  vector<int> close_pt(lM,-1) ; 

  MPI_Barrier(MPI_COMM_WORLD) ;

  get_timer() ;

  //---------------------------------------------------------------------
  // Insert your parallel algorithm here
  //
  
  // (This is just the serial version as a placeholder)
  kd_tree kd(target_pnts,target_ids) ;
  for(int i=0;i<lM;++i) {
    close_pt[i] = kd.find_closest(search_pnts[i]) ;
  }
    
  //
  // End of parallel nearest neighbor algorithm
  //---------------------------------------------------------------------

  double t = get_timer() ;
  double tmax = t ;
  MPI_Reduce(&t,&tmax,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD) ;
  if(0 == myid) {
    cout << "time to complete parallel nearest neighbor search is "
         <<  tmax << endl ;
  }

  if(verify) {
    if(0==myid) 
      cout << "verifying result...." << endl ;
    Verify(target_pnts,target_ids,search_pnts,close_pt) ;
  }

  MPI_Finalize() ;
  return 0 ;
}
